#define _CRT_SECURE_NO_WARNINGS
#define _CRT_NONSTDC_NO_DEPRECATE

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include <time.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include "mwis.h"

unsigned char* mwisRandAdjMat( unsigned char* A, const long n, const long ne)
{
	long i,j;
	long cnt = 0,cnt2 = n*n-n;
	unsigned char c;

	if(A<=NULL)A=(unsigned char *)malloc(n*n*sizeof(unsigned char));

	memset(A,0,n*n*sizeof(unsigned char));
	
	srand((unsigned int)time(0));
	for(i=0;i<n;i++)
	{
		for(j=i+1;j<n;j++)
		{
			c = cnt2 < ne-cnt?1:cnt>ne?0:rand()%2 ;
			cnt += c;
			A[i*n+j] = A[j*n+i] = c;
			cnt2--;
		}
	}

	return A;
}

MAT_TYPE* mwisRandWeight( MAT_TYPE* w, const long n)
{
	long i;

	if(w<=NULL)w=(MAT_TYPE *)malloc(n*sizeof(MAT_TYPE));
	
	srand((unsigned int)time(0));
	for(i=0;i<n;i++)
	{
		//w[i] = (MAT_TYPE)(rand())/(MAT_TYPE)RAND_MAX;
		w[i] = (MAT_TYPE)(rand()%10)/(MAT_TYPE)10;
	}

	return w;
}

void mwisRandGraph( unsigned char** A,MAT_TYPE** w, const long n, const long ne)
{
	*w = mwisRandWeight( *w, n);
	*A = mwisRandAdjMat( *A, n, ne);
}

void mwisPrintStats( const MAT_TYPE* w, const unsigned char* A, const long n, const MAT_TYPE alpha,const MAT_TYPE lmin, const unsigned char*x )
{
	long i,j;

	/* compute xtAx and wtx*/
	long xtAx = 0;
	double wtx = 0;

	for(i=0;i<n;i++)
	{
		wtx+= w[i]*x[i];
		for(j=0;j<n;j++) xtAx+= x[i]*x[j]*A[i*n+j];
	}

	printf("objective: %g	constraints: %d alpha: %g lmin: %d\n",wtx-alpha*0.5*xtAx,xtAx,alpha,lmin);
}

void mwisPrintAdjMat(const unsigned char* A, const long n)
{
	long i,j;
	for(i=0;i<n;i++)
	{
		for(j=0;j<n;j++)
		{
			printf("%s%d",j?" ":"",A[i*n+j]);
		}
		printf("\n");
	}
}

void mwisPrintWeights(const MAT_TYPE* w, const long n)
{
	long i;
	for(i=0;i<n;i++)
	{
		printf("%g\n",w[i]);
	}
}

void mwisPrintX(const unsigned char* x, const long n)
{
	long i;
	for(i=0;i<n;i++)
	{
		printf("%d\n",x[i]);
	}
}

int mwisRescaleWeights( MAT_TYPE* w, const long n)
{
	long i,j=0;
	MAT_TYPE* pt = w;
	MAT_TYPE maxw = -1e20,minw = 1e20,rangew;

	/* init y,x and rescale w */
	for(i=0;i<n;i++,pt++) 
	{
		if(maxw < *pt){ maxw = *pt;j=i;}
		if(minw > *pt) minw = *pt;
	}
	rangew = maxw - minw;
	if(rangew<=0) return j;

	for(i=n,pt=w;i--;pt++)
	{
		*pt = (*pt-minw)/rangew;
	}

	return j;
}

MAT_TYPE mwisxtAx(const MAT_TYPE* Ax, const long n, const MAT_TYPE* x)
{
	long i = n;
	MAT_TYPE res = 0;
	MAT_TYPE* ptx = (MAT_TYPE*)x;
	MAT_TYPE* ptAx = (MAT_TYPE*)Ax;
	while(i--) 
		res += *(ptAx++) * *(ptx++);

	return res;
}

MAT_TYPE* mwisAx(const unsigned char* A, const long n, const MAT_TYPE* x, MAT_TYPE* Ax)
{
	long i = n,j;
	unsigned char* ptA = (unsigned char*)A;
	MAT_TYPE* ptAx = Ax;
	MAT_TYPE* ptx = (MAT_TYPE*)x;
	
	memset(Ax,0,n*sizeof(MAT_TYPE));

	for(;i--;ptAx++)
	{
		for(j=0;j<n;j++,ptA++) (*ptAx) += *ptA * x[j];
	}

	return Ax;
}

MAT_TYPE mwiswtx( const MAT_TYPE* w, const long n, const MAT_TYPE* x)
{
	long i = n;
	MAT_TYPE* ptx = (MAT_TYPE*)x;
	MAT_TYPE* ptw = (MAT_TYPE*)w;
	MAT_TYPE res = 0;

	while(i--) 
		res += *(ptw++) * *(ptx++);

	return res;
}

MAT_TYPE mwisxtx( const MAT_TYPE* x, const long n)
{
	long i = n;
	MAT_TYPE* ptx = (MAT_TYPE*)x;
	MAT_TYPE res = 0;

	while(i--) 
		res += *(ptx) * *(ptx++);

	return res;
}

MAT_TYPE mwisObjFuncX( const MAT_TYPE* w, const MAT_TYPE* Ax, const long n, 
					   const MAT_TYPE* x , const MAT_TYPE alpha)
{
	return mwiswtx( w, n, x) - alpha*mwisxtAx(Ax, n, x)/2;
}

void mwisApproxX( const MAT_TYPE* w, const unsigned char* A, const long n, 
				  MAT_TYPE* xtilde, 
				  const MAT_TYPE alpha,
				  const MAT_TYPE* Axprev)
{
	long i,maxi=-1;
	MAT_TYPE maxval=-1e20,val;
	MAT_TYPE* ptAxprev = (MAT_TYPE*)Axprev;
	MAT_TYPE *ptw = (MAT_TYPE*)w;
	unsigned char flag = 0;
	MAT_TYPE *ptxtilde = xtilde;

	for(i=0;i<n;i++)
	{
		val = *(ptw++) - alpha * *(ptAxprev++);
		flag |= (val>0);
		if(maxval<val)
		{
			maxi = i;
			maxval = val;
		}
		*(ptxtilde++) = (val>0);
	}
	if(!flag) 
		xtilde[maxi] = 1;
}
void mwisApproxX1( const MAT_TYPE* w, const unsigned char* A, const long n, 
				  const MAT_TYPE* xprev , MAT_TYPE* xtilde, 
				  const MAT_TYPE alpha, const long lmin,
				  const MAT_TYPE* Axprev)
{
	long i,maxi=-1;
	MAT_TYPE maxval=-1e20,val;
	MAT_TYPE* ptAxprev = (MAT_TYPE*)Axprev;
	MAT_TYPE *ptw = (MAT_TYPE*)w;
	unsigned char flag = 0;
	MAT_TYPE *ptxtilde = xtilde;
	MAT_TYPE *ptxprev = (MAT_TYPE*)xprev;
	MAT_TYPE alphaLmin_2 = alpha*lmin*0.5;

	for(i=0;i<n;i++)
	{
		val = *(ptw++) - alphaLmin_2 - alpha * ( *(ptAxprev++)- lmin * *(ptxprev++) );
		flag |= (val>0);
		if(maxval<val)
		{
			maxi = i;
			maxval = val;
		}
		*(ptxtilde++) = (val>0);
	}
	if(!flag) 
		xtilde[maxi] = 1;
}

void mwisApproxX2( const MAT_TYPE* w, const unsigned char* A, const long n, 
				  const MAT_TYPE* xprev , MAT_TYPE* xtilde, 
				  const MAT_TYPE alpha, const MAT_TYPE eps,
				  const MAT_TYPE* Axprev)
{
	long i,maxi=-1;
	MAT_TYPE maxval=-1e20,val;
	MAT_TYPE* ptAxprev = (MAT_TYPE*)Axprev;
	MAT_TYPE *ptw = (MAT_TYPE*)w;
	unsigned char flag = 0;
	MAT_TYPE *ptxtilde = xtilde;
	MAT_TYPE *ptxprev = (MAT_TYPE*)xprev;
	MAT_TYPE alphaEps_2 = alpha*eps*0.5;

	for(i=0;i<n;i++)
	{
		val = *(ptw++) + alphaEps_2 - alpha * ( *(ptAxprev++) + eps * *(ptxprev++) );
		flag |= (val>0);
		if(maxval<val)
		{
			maxi = i;
			maxval = val;
		}
		*(ptxtilde++) = (val>0);
	}
	if(!flag)
		xtilde[maxi] = 1;
}

void mwisApproxX3( const MAT_TYPE* w, const unsigned char* A, const long n, 
				  const MAT_TYPE* xprev , MAT_TYPE* xtilde, 
				  const MAT_TYPE alpha, const MAT_TYPE eps, const long lmin,
				  const MAT_TYPE* Axprev)
{
	long i,maxi=-1;
	MAT_TYPE maxval=-1e20,val;
	MAT_TYPE* ptAxprev = (MAT_TYPE*)Axprev;
	MAT_TYPE *ptw = (MAT_TYPE*)w;
	unsigned char flag = 0;
	MAT_TYPE *ptxtilde = xtilde;
	MAT_TYPE *ptxprev = (MAT_TYPE*)xprev;
	MAT_TYPE alphaEps_2 = alpha*eps*0.5;
	MAT_TYPE alphaLmin_2 = alpha*lmin*0.5;

	for(i=0;i<n;i++)
	{
		val = *(ptw++) + alphaEps_2 - alphaLmin_2 - alpha * ( *(ptAxprev++) + (eps-lmin) * *(ptxprev++) );
		flag |= (val>0);
		if(maxval<val)
		{
			maxi = i;
			maxval = val;
		}
		*(ptxtilde++) = (val>0);
	}
	if(!flag)
		xtilde[maxi] = 1;
}

MAT_TYPE mwisComputeEps(const MAT_TYPE* Axtilde, const MAT_TYPE* xtilde,
						const MAT_TYPE* Axprev, const MAT_TYPE* xprev,
						const long n)
{
	long i=n;
	MAT_TYPE c=0, d= 0, val;
	MAT_TYPE* ptxtilde = (MAT_TYPE*)xtilde;
	MAT_TYPE* ptxprev = (MAT_TYPE*)xprev;
	MAT_TYPE* ptAxtilde = (MAT_TYPE*)Axtilde;
	MAT_TYPE* ptAxprev = (MAT_TYPE*)Axprev;

	while(i--)
	{
		val = *(ptxtilde++) - *(ptxprev++);
		c += (*(ptAxtilde++) - *(ptAxprev++)) * val;
		d +=  val*val;
	}

	return d!=0? fabs(c/d):0;
}

MAT_TYPE mwisRefineAlpha(const MAT_TYPE* xtilde, const MAT_TYPE* Axtilde, const long n, const MAT_TYPE alpha)
{
	long i = n;
	MAT_TYPE res = 0;
	MAT_TYPE* ptx = (MAT_TYPE*)xtilde;
	MAT_TYPE* ptAx = (MAT_TYPE*)Axtilde;
	while(i--) 
		res += *(ptAx++) * *(ptx++);

	/*if(res<0)
	{
		printf("aaaaa\n");
		for(i=0;i<n;i++)
			if(xtilde[i] != (int)xtilde[i])
				printf("%g\n",xtilde[i]);
	}*/
	//res = max(res,0);
	
	return alpha + 0.5*res;
}

MAT_TYPE mwisRefineX(const MAT_TYPE* w, const unsigned char* A, const long n,
					 MAT_TYPE* Axtilde, MAT_TYPE* xtilde,
					 const MAT_TYPE* Axprev, const MAT_TYPE* xprev,
					 MAT_TYPE alpha)
{
	long i=n;
	MAT_TYPE c,d,eta,_1_eta;
	MAT_TYPE val;
	MAT_TYPE* ptxtilde = xtilde;
	MAT_TYPE* ptxprev = (MAT_TYPE*)xprev;
	MAT_TYPE* ptAxtilde = Axtilde;
	MAT_TYPE* ptAxprev = (MAT_TYPE*)Axprev;
	MAT_TYPE* ptw = (MAT_TYPE*)w;

	/* compute eta */
	c = d = 0;

	while(i--)
	{
		val = *(ptxtilde++) - *(ptxprev++);
		c += (*(ptw++) - alpha * *ptAxprev) * val;
		d += (*(ptAxtilde++) - *(ptAxprev++)) * val;
	}
	d *= alpha;

	eta = d!=0? c/d:0;
	eta = eta>1?1:eta<0?0:eta;
	_1_eta = 1-eta;

	/* compute xtilde */
	i=n;
	ptxtilde = (MAT_TYPE*)xtilde;
	ptxprev = (MAT_TYPE*)xprev;
	for(;i--;ptxtilde++) *ptxtilde = eta * *ptxtilde + _1_eta * *(ptxprev++);

	/* compute objective function with xtilde */
	return mwisObjFuncX( w, mwisAx(A, n, xtilde, Axtilde), n, xtilde, alpha);
}

long mwisLmin(const unsigned char* A, const long n)
{
	long i=n,j;
	unsigned char* pt = (unsigned char*)A;
	long lmin = 0,l;

	while(i--)
	{
		l = 2*A[i*n+i];
		j=n;
		while(j--)
		{
			l-= *(pt++);
		}
		lmin = min(lmin,l);
	}

	return lmin;
}

unsigned char* mwis2( MAT_TYPE* w, unsigned char* A, long n, unsigned char* x, unsigned char flag)
{
	long i,j=0;
	int maxIter			= 100;
	MAT_TYPE alpha		= 2;
	MAT_TYPE norm2		= 1e20;
	MAT_TYPE threshold	= 1e-16;

	MAT_TYPE objf,objftemp, val, eps = 0;
	void* temp;
	MAT_TYPE* xtilde;
	MAT_TYPE* xprev;
	MAT_TYPE* Axtilde;
	MAT_TYPE* Axprev;
	long lmin = flag?mwisLmin(A, n):0;

	/* rescale w */
	j = mwisRescaleWeights( w, n);

	/* allocate memory */
	if(x<=NULL) x = (unsigned char*)malloc(n*sizeof(unsigned char));

	xtilde	= (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	Axtilde	= (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	xprev	= (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	Axprev	= (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));

	/* init xprev and Axprev */
	memset(xprev,0,n*sizeof(MAT_TYPE));
	x[j] = (unsigned char)(xprev[j] = 1);
	if(flag==1) for(i=0;i<n;i++) x[i] = (unsigned char)(xprev[i] = (rand()%2));
	if(flag==2) for(i=0;i<n;i++) xprev[i] = w[i];
	if(flag>2) for(i=0;i<n;i++) xprev[i] = rand()%2?w[i]:(rand()%2);
	mwisAx(A, n, xprev, Axprev);

	/* compute objective function */
	objf = mwisObjFuncX( w, Axprev, n, xprev, alpha);

	//printf("init objf: %g\n",objf);
	//mwisPrintWeights(xprev,n);

	/* iterate */
	for(;(maxIter--) &&  norm2 >threshold; )
	{
		/* find local solution xtilde */
		mwisApproxX( w, A, n, xtilde, alpha, Axprev);
		//mwisApproxX1( w,  A, n, xprev , xtilde, alpha,lmin, Axprev);
		//mwisApproxX2( w,  A, n, xprev , xtilde, alpha,eps, Axprev);
		//mwisApproxX3( w,  A, n, xprev , xtilde, alpha,eps,lmin, Axprev);

		/* compute objective function with xtilde */
		objftemp = mwisObjFuncX( w, mwisAx(A, n, xtilde, Axtilde), n, xtilde, alpha);

		/* check if objective function increases */
		if(objftemp>=objf)
		{
			/* compute norm and update x*/
			norm2 = 0;
			for(i=0;i<n;i++) 
			{
				val = xprev[i] - (MAT_TYPE)(x[i] = (unsigned char)xtilde[i]);
				norm2 += val*val;
			}

			objf = objftemp;
		}
		else
		{
			//printf("objftemp: %g < objf: %g\n",objftemp,objf);
			objf =  mwisRefineX(w, A, n, Axtilde,  xtilde, Axprev, xprev, alpha);
			
			/* compute norm */
			norm2 = 0;
			for(i=0;i<n;i++) 
			{
				val = xprev[i] - xtilde[i];
				norm2 += val*val;
			}
		}

		//printf("> objf: %g ",objf);
		//mwisPrintStats( w, A, n, alpha,lmin, x );

		if(flag<4)alpha = mwisRefineAlpha(xtilde, Axtilde, n, alpha);
		eps = flag?mwisComputeEps(Axtilde, xtilde, Axprev, xprev, n):0;
		//eps = -lmin;

		/* swap memory */
		temp = (void*) xprev;
		xprev = xtilde;
		xtilde = (MAT_TYPE*)temp;
		temp = (void*) Axprev;
		Axprev = Axtilde;
		Axtilde = (MAT_TYPE*)temp;
	}

	printf("%d ",maxIter);
	mwisPrintStats( w, A, n, alpha,lmin, x );

	/* free memory */
	free(xtilde);
	free(Axtilde);
	free(xprev);
	free(Axprev);

	return x;
}
unsigned char* mwis( MAT_TYPE* w, unsigned char* A, long n, unsigned char* x)
{
	long i,j=0;
	int maxIter			= 100;
	MAT_TYPE alpha		= 2;
	MAT_TYPE norm2		= 1e20;
	MAT_TYPE threshold	= 1e-16;

	MAT_TYPE objf,objftemp, val;
	void* temp;
	MAT_TYPE* xtilde;
	MAT_TYPE* xprev;
	MAT_TYPE* Axtilde;
	MAT_TYPE* Axprev;

	/* rescale w */
	j = mwisRescaleWeights( w, n);

	/* allocate memory */
	if(x<=NULL) x = (unsigned char*)malloc(n*sizeof(unsigned char));

	xtilde	= (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	Axtilde	= (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	xprev	= (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	Axprev	= (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));

	/* init xprev and Axprev */
	memset(xprev,0,n*sizeof(MAT_TYPE));
	x[j] = (unsigned char)(xprev[j] = 1);
	//for(i=0;i<n;i++) x[i] = (unsigned char)(xprev[i] = (rand()%2));
	for(i=0;i<n;i++) xprev[i] = w[i];
	//for(i=0;i<n;i++) xprev[i] = rand()%2?w[i]:(rand()%2);
	mwisAx(A, n, xprev, Axprev);

	/* compute objective function */
	objf = mwisObjFuncX( w, Axprev, n, xprev, alpha);

	/* iterate */
	for(;(maxIter--) &&  norm2 >threshold; )
	{
		/* find local solution xtilde */
		mwisApproxX( w, A, n, xtilde, alpha, Axprev);

		/* compute objective function with xtilde */
		objftemp = mwisObjFuncX( w, mwisAx(A, n, xtilde, Axtilde), n, xtilde, alpha);

		/* check if objective function increases */
		if(objftemp>=objf)
		{
			/* compute norm and update x */
			norm2 = 0;
			for(i=0;i<n;i++) 
			{
				val = xprev[i] - (MAT_TYPE)(x[i] = (unsigned char)xtilde[i]);
				norm2 += val*val;
			}

			objf = objftemp;
		}
		else
		{
			/* compute best point by linear interpolation */
			objf =  mwisRefineX(w, A, n, Axtilde,  xtilde, Axprev, xprev, alpha);
			
			/* compute norm */
			norm2 = 0;
			for(i=0;i<n;i++) 
			{
				val = xprev[i] - xtilde[i];
				norm2 += val*val;
			}
		}

		/* swap memory */
		temp = (void*) xprev;
		xprev = xtilde;
		xtilde = (MAT_TYPE*)temp;
		temp = (void*) Axprev;
		Axprev = Axtilde;
		Axtilde = (MAT_TYPE*)temp;
	}

	printf("%d ",maxIter);
	mwisPrintStats( w, A, n, alpha,0, x );

	/* free memory */
	free(xtilde);
	free(Axtilde);
	free(xprev);
	free(Axprev);

	return x;
}