#include "MWIS.h"

#define alpha  2

MAT_TYPE mwisComputeObjFuncX( MAT_TYPE* w, MAT_TYPE* A, int n, unsigned char* x)
{
	int i,j;
	MAT_TYPE goal = 0,val = 0;

	for(i=0;i<n;i++)
		for(j=0;j<n;j++)
			val += A[i*n+j]*x[i]*x[j];
	goal = - 0.5*alpha* val;

	for(i=0;i<n;i++)
		goal += w[i]*x[i];

	return goal;
}

MAT_TYPE mwisComputeObjFuncY( MAT_TYPE* w, MAT_TYPE* A, int n, MAT_TYPE* y)
{
	int i,j;
	MAT_TYPE goal = 0,val = 0;

	for(i=0;i<n;i++)
		for(j=0;j<n;j++)
			val += A[i*n+j]*y[i]*y[j];
	goal = - 0.5*alpha* val;

	for(i=0;i<n;i++)
		goal += w[i]*y[i];

	return goal;
}

MAT_TYPE mwisComputeConvexObjFuncY( MAT_TYPE* w, MAT_TYPE c, int n, MAT_TYPE* y)
{
	int i;
	MAT_TYPE goal = 0,val = 0;

	for(i=0;i<n;i++)
		val += y[i]*y[i];
	goal = - 0.5*c*alpha* val;

	for(i=0;i<n;i++)
		goal += w[i]*y[i];

	return goal;
}

MAT_TYPE mwisComputeConvexObjFuncX( MAT_TYPE* w, MAT_TYPE c, int n, unsigned char* x)
{
	int i;
	MAT_TYPE goal = 0,val = 0;

	for(i=0;i<n;i++)
		val += x[i]*x[i];
	goal = - 0.5*c*alpha* val;

	for(i=0;i<n;i++)
		goal += w[i]*x[i];

	return goal;
}

MAT_TYPE mwisConstraint(MAT_TYPE* A, int n, MAT_TYPE* y)
{
	int i,j;
	MAT_TYPE val = 0;

	for(i=0;i<n;i++)
		for(j=0;j<n;j++)
			val += A[i*n+j]*y[i]*y[j];
	return val;
}

/* MWIS based on integer fixed point */
unsigned char* mwisGetX1( MAT_TYPE* w, MAT_TYPE* A, int n, unsigned char* x)
{
	int i,j,maxIter = 1000;
	MAT_TYPE maxw = -1e20,minw = 1e20,rangew;
	MAT_TYPE norm2 = 1e20;
	MAT_TYPE threshold = 1e-16;
	MAT_TYPE eta;
	MAT_TYPE c,d,val,maxval;
	int maxi;
	unsigned char flag;

	MAT_TYPE objf,objftemp;

	MAT_TYPE* z = (char*)malloc(n*sizeof(char));
	MAT_TYPE* y = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	unsigned char* xtilde = (unsigned char*)malloc(n*sizeof(unsigned char));

	/* allocate memory if needed */
	if(x<=NULL) x = (unsigned char*)malloc(n*sizeof(unsigned char));

	/* init y,x and rescale w */
	for(i=0;i<n;i++) 
	{
		if(maxw < w[i]){ maxw = w[i];j=i;}
		if(minw > w[i]) minw = w[i];
	}
	rangew = maxw - minw;
	srand((unsigned int)time(0));
	for(i=0;i<n;i++)
	{
		w[i] = (w[i]-minw)/rangew;
		y[i] = /*w[i]*(rand()%2);*/0.5*w[i] + 0.01;/**/
	}

	memset(x,0,n*sizeof(unsigned char));
	x[j] = 1;
	//memset(y,0,n*sizeof(MAT_TYPE));
	//y[j] = 1;

	/* compute objective function */
	objf = mwisComputeObjFuncY( w, A, n, y);
	printf("objf: %f\n",objf);

	/* iterate */
	for(;(maxIter--) &&  norm2 >threshold; )
	{
		//printf("\niter: %d\n",maxIter);
		/* find local solution xtilde */
		maxval = -1e20;
		maxi = -1;
		flag = 0;
		for(i=0;i<n;i++)
		{
			val = 0;
			for(j=0;j<n;j++)
			{
				val += A[i*n+j]*y[j];
			}
			val = w[i] - alpha*val;
			xtilde[i] = (val>=0);
			flag |= xtilde[i];

			if(maxval<val)
			{
				maxval = val;
				maxi = i;
			}
		}

		if(!flag) xtilde[maxi] = 1;

		/* compute objective function with xtilde */
		objftemp = mwisComputeObjFuncX( w, A, n, xtilde);

		/* check if objective function increases */
		if(objftemp>=objf)
		{
			/* compute norm */
			norm2 = 0;
			for(i=0;i<n;i++) norm2 += (y[i] - xtilde[i])*(y[i] - xtilde[i]);

			/* update y and x */
			for(i=0;i<n;i++)
			{
				//if(xtilde[i])
				//	printf("i: %d y: %lf xtilde: %d x: %d\n",i,y[i],xtilde[i],x[i]);
				y[i] = xtilde[i];
			}
			memcpy(x,xtilde,n*sizeof(unsigned char));
			objf = objftemp;
		}
		else
		{
			/* compute z = xtilde - y */
			for(i=0;i<n;i++) z[i] = xtilde[i] - y[i];

			/* compute eta */
			c = d = 0;

			/* (w - alpha*Ay)z */
			for(i=0;i<n;i++)
			{
				val = 0;
				for(j=0;j<n;j++)
				{
					val += A[i*n+j]*y[j];
				}
				c += (w[i] - alpha*val)*z[i];
			}

			/* zAz */
			for(i=0;i<n;i++)
				for(j=0;j<n;j++)
					d += A[i*n+j]*z[i]*z[j];

			eta = d>0? max(c/(alpha*d),0):min(c/(alpha*d),1);

			/* compute norm */
			norm2 = 0;
			for(i=0;i<n;i++) norm2 += eta*z[i]*eta*z[i];

			/* compute y */
			for(i=0;i<n;i++) y[i] += eta*z[i];

			/* compute objective function with y */
			objf = mwisComputeObjFuncY( w, A, n, y);
		}
		//printf("objf: %f\n",objf);
	}

	printf("objf: %f\n",objf);

	/* free memory */
	free(z);
	free(y);
	free(xtilde);

	return x;
}

/* MWIS based on gradient descent */
unsigned char* mwisGetX2( MAT_TYPE* w, MAT_TYPE* A, int n, unsigned char* x)
{
	int i,j,maxIter = 1000;
	MAT_TYPE maxw = -1e20,minw = 1e20,rangew;
	MAT_TYPE norm2 = 1e20;
	MAT_TYPE threshold = 1e-16;
	MAT_TYPE eta;
	MAT_TYPE c,d,val,dt;
	MAT_TYPE objf,objftemp,objfX;

	MAT_TYPE* ytemp;
	MAT_TYPE* y = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	MAT_TYPE* dy = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	MAT_TYPE* ynext = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	MAT_TYPE* z = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));

	/* allocate memory if needed */
	if(x<=NULL) x = (unsigned char*)malloc(n*sizeof(unsigned char));

	/* init y,x, and rescale w */
	for(i=0;i<n;i++) 
	{
		if(maxw < w[i]){ maxw = w[i];j=i;}
		if(minw > w[i]) minw = w[i];
	}
	rangew = maxw - minw;
	srand((unsigned int)time(0));
	for(i=0;i<n;i++)
	{
		w[i] = (w[i]-minw)/rangew;
		y[i] = w[i]*(rand()%2)/**/;/**//*0.5*w[i] + 0.01;*/
	}
	memset(x,0,n*sizeof(unsigned char));
	x[j] = 1;

	/* compute objective function */
	objf = mwisComputeObjFuncY( w, A, n, y);
	printf("objf: %f\n",objf);

	/* iterate */
	for(;(maxIter--) &&  norm2 >threshold; )
	{
		/* find gradient dy */
		for(i=0;i<n;i++)
		{
			val = 0;
			for(j=0;j<n;j++)
			{
				val += A[i*n+j]*y[j];
			}
			dy[i] = w[i] - alpha*val;
		}

		/* find time step dt */
		dt = 0;
		for(i=0;i<n;i++) dt+= dy[i]*dy[i];
		val = 0;
		for(i=0;i<n;i++)
			for(j=0;j<n;j++)
				val += A[i*n+j]*dy[i]*dy[j];
		if(val) 
			dt = dt/(alpha*fabs(val));

		/* find local solution ynext */
		for(i=0;i<n;i++) 
		{
			ynext[i] = y[i] + dt*dy[i];
			if(ynext[i]>1) ynext[i] = 1;
			if(ynext[i]<0) ynext[i] = 0;
		}

		/* update x and compute objective function with x */
		for(i=0;i<n;i++)x[i] = (ynext[i]>1e-16);
		objfX = mwisComputeObjFuncX( w, A, n, x);

		/* compute objective function with ynext */
		objftemp = mwisComputeObjFuncY( w, A, n, ynext);

		if(objfX>objftemp)
		{
			for(i=0;i<n;i++) ynext[i] = x[i];
			objftemp = objfX;
		}

		if(objftemp<objf)
		{
			/* compute z = ynext - y */
			for(i=0;i<n;i++) z[i] = ynext[i] - y[i];
			
			c = d = 0;

			/* (w - alpha*Ay)z */
			for(i=0;i<n;i++)
				c += dy[i]*z[i];

			/* zAz */
			for(i=0;i<n;i++)
				for(j=0;j<n;j++)
					d += A[i*n+j]*z[i]*z[j];

			/* compute eta */
			eta = d>0? max(c/(alpha*d),0):min(c/(alpha*d),1);

			/* compute ynext */
			for(i=0;i<n;i++) ynext[i] = y[i] + eta*z[i];

			/* compute objective function with ynext */
			objftemp = mwisComputeObjFuncY( w, A, n, ynext);
		}

		/* compute norm */
		norm2 = 0;
		/*for(i=0;i<n;i++) norm2 += (y[i] - ynext[i])*(y[i] - ynext[i]);
		norm2/=n;*/
		for(i=0;i<n;i++) norm2 = max(norm2,fabs(y[i] - ynext[i]));

		/* info */
		//printf("objf: %f norm2: %g dt: %g iter: %d yAy: %g\n",objf,norm2,dt,maxIter,mwisConstraint(A, n, y));

		/* update for next iteration */
		objf = objftemp;
		ytemp = y;
		y = ynext;
		ynext = ytemp;
	}

	/* update x */
	for(i=0;i<n;i++)x[i] = (y[i]>=1e-16);

	/* info */
	printf("objf: %f norm2: %g dt: %g iter: %d yAy: %g\n",objf,norm2,dt,maxIter,mwisConstraint(A, n, y));

	/* free memory */
	free(y);
	free(ynext);
	free(dy);
	free(z);

	return x;
}
/* MWIS based on polynimial formulation */
unsigned char* mwisGetX3( MAT_TYPE* w, MAT_TYPE* A, int n, unsigned char* x)
{
	int i,j,maxIter = 1000;
	MAT_TYPE maxw = -1e20,minw = 1e20,rangew;
	MAT_TYPE norm2 = 1e20;
	MAT_TYPE threshold = 1e-16;
	MAT_TYPE beta = 50;
	MAT_TYPE dt=100;
	MAT_TYPE objf,objfmax,dtmax;

	MAT_TYPE* temp;
	MAT_TYPE* _x = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	MAT_TYPE* y = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	MAT_TYPE* h = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	MAT_TYPE* ynext = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	MAT_TYPE* _xnext = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	MAT_TYPE* hnext = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));

	/* allocate memory if needed */
	if(x<=NULL) x = (unsigned char*)malloc(n*sizeof(unsigned char));

	/* init y,x and h and rescale w */
	for(i=0;i<n;i++) 
	{
		if(maxw < w[i]){ maxw = w[i];j=i;}
		if(minw > w[i]) minw = w[i];
	}
	rangew = maxw - minw;
	srand((unsigned int)time(0));
	for(i=0;i<n;i++)
	{
		w[i] = (w[i]-minw)/rangew;
		y[i] = w[i]*(rand()%2)/**/;/**//*0.5*w[i] + 0.01;*/
		_x[i] = 1./(1+exp(-beta*(y[i]-0.5)));
	}

	objf = 0;
	for(i=0;i<n;i++)
	{
		h[i] = _x[i];
		for(j=0;j<n;j++) if(A[i*n+j]>0) h[i]*= 1-_x[j];
		objf += h[i]*w[i];
	}
	memset(x,0,n*sizeof(unsigned char));
	x[j] = 1;

	/* objective function */
	printf("objf: %f\n",objf);

	/* iterate */
	for(;(maxIter--) &&  norm2 >threshold; )
	{
		/* line search */
		objfmax=0; dtmax = 1;
		for(dt=0.1;dt<=1000;dt*=2)
		{
			/* update y */
			for(i=0;i<n;i++)
			{
				ynext[i] = y[i] + dt*(1-_x[i])*h[i]*w[i];
				for(j=0;j<n;j++) if(A[i*n+j]>0) ynext[i] -= dt*_x[j]*h[j]*w[j];
			}

			/* update x */
			for(i=0;i<n;i++)
				_xnext[i] = 1./(1+exp(-beta*(ynext[i]-0.5)));

			/* update h */
			objf = 0;
			for(i=0;i<n;i++)
			{
				hnext[i] = _xnext[i];
				for(j=0;j<n;j++) if(A[i*n+j]>0) hnext[i]*= 1-_xnext[j];
				objf += hnext[i]*w[i];
			}

			if(objfmax<objf)
			{
				objfmax = objf;
				dtmax = dt;
			}
		}

		/* update y */
		dt = dtmax; objf= objfmax;
		for(i=0;i<n;i++)
		{
			ynext[i] = y[i] + dt*(1-_x[i])*h[i]*w[i];
			for(j=0;j<n;j++) if(A[i*n+j]>0) ynext[i] -= dt*_x[j]*h[j]*w[j];
		}

		/* update x */
		for(i=0;i<n;i++)
			_xnext[i] = 1./(1+exp(-beta*(ynext[i]-0.5)));

		/* update h */
		for(i=0;i<n;i++)
		{
			hnext[i] = _xnext[i];
			for(j=0;j<n;j++) if(A[i*n+j]>0) hnext[i]*= 1-_xnext[j];
		}

		/* norm */
		norm2 = 0;
		//for(i=0;i<n;i++) norm2 += (_xnext[i]-_x[i])*(_xnext[i]-_x[i]);
		//norm2/=n;
		//for(i=0;i<n;i++) if( fabs((int)(_xnext[i] +0.5)-_xnext[i])>0.1 ) norm2=1;
		for(i=0;i<n;i++) norm2 += (ynext[i]-y[i])*(ynext[i]-y[i]);
		norm2/=n*dt*dt;

		/* objective function */
		printf("objf: %f norm2: %g dt: %f\n",objf,norm2,dt);


		temp = ynext;
		ynext = y;
		y = temp;

		temp = _xnext;
		_xnext = _x;
		_x = temp;

		temp = hnext;
		hnext = h;
		h = temp;
	}

	/* update x */
	for(i=0;i<n;i++) x[i] = (_x[i]>0.4);

	/* objective function */
	printf("objf: %f\n",objf);

	/* free memory */
	free(y);
	free(ynext);
	free(_x);
	free(_xnext);
	free(h);
	free(hnext);

	return x;
}

/* MWIS based on dual decomposition */
unsigned char* mwisGetX4( MAT_TYPE* w, MAT_TYPE* A, int n, unsigned char* x)
{
	int i,j,maxIter = 1000;
	MAT_TYPE maxw = -1e20,minw = 1e20,rangew;
	MAT_TYPE norm2 = 1e20;
	MAT_TYPE threshold = 1e-16;
	unsigned char** xi;
	MAT_TYPE** li;
	MAT_TYPE* avgli,*y,*ynext,*temp;
	unsigned char* x0;
	unsigned char* x1;
	MAT_TYPE h0,h1,val,dt = 0.1*(1-1./n);

	/* allocate memory if needed */
	if(x<=NULL) x = (unsigned char*)malloc(n*sizeof(unsigned char));
	xi = (unsigned char**)malloc(n*sizeof(unsigned char*));
	*xi = (unsigned char*)malloc(n*n*sizeof(unsigned char));
	li = (MAT_TYPE**)malloc(n*sizeof(MAT_TYPE*));
	*li = (MAT_TYPE*)malloc(n*n*sizeof(MAT_TYPE));
	memset(*li,0,n*n*sizeof(MAT_TYPE));
	avgli = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	x0 = (unsigned char*)malloc(n*sizeof(unsigned char));
	x1 = (unsigned char*)malloc(n*sizeof(unsigned char));
	y = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));
	ynext = (MAT_TYPE*)malloc(n*sizeof(MAT_TYPE));

	for(i=0;i<n;i++)
	{
		xi[i] = *xi + i*n;
		li[i] = *li + i*n;
	}

	/* init x and rescale w */
	for(i=0;i<n;i++) 
	{
		if(maxw < w[i]){ maxw = w[i];j=i;}
		if(minw > w[i]) minw = w[i];
	}
	rangew = maxw - minw;
	for(i=0;i<n;i++)
		w[i] = (w[i]-minw)/rangew;
	memset(x,0,n*sizeof(unsigned char));
	x[j] = 1;
	memset(y,0,n*sizeof(MAT_TYPE));
	y[j] = 1;
	memset(avgli,0,n*sizeof(MAT_TYPE));

	/* algorithm */
	for(;(maxIter--) &&  norm2 > threshold; )
	{
		/* update each xi */
		for(i=0;i<n;i++)
		{
			/* compute x0 */
			h0 = 0;
			for(j=0;j<n;j++)
			{
				val = li[i][j]/*-avgli[j]*/;
				x0[j] = (val>=0 && i!=j);
				h0 += val*x0[j];
			}

			/* compute x1 */
			h1 = w[i];
			for(j=0;j<n;j++)
			{
				val = li[i][j]/*-avgli[j]*/-alpha/2.*A[i*n+j];
				x1[j] = (val>=0 || i==j);
				h1 += val*x1[j];
			}

			if(h0>h1)	memcpy(xi[i],x0,n*sizeof(unsigned char));
			else		memcpy(xi[i],x1,n*sizeof(unsigned char));

		}

		/* update li*/
		for(i=0;i<n;i++)
			for(j=0;j<n;j++)
				li[i][j] -= dt*xi[i][j];

		/* update average of li */
		memset(avgli,0,n*sizeof(MAT_TYPE));
		for(i=0;i<n;i++)
			for(j=0;j<n;j++)
				avgli[j] += li[i][j];
		for(i=0;i<n;i++)
			avgli[i]/=n;
		
		/* update li */
		for(i=0;i<n;i++)
		{
			for(j=0;j<n;j++)
			{
				li[i][j] -= avgli[j];
			}
		}/**/

		/* update y */
		memset(ynext,0,n*sizeof(MAT_TYPE));
		for(i=0;i<n;i++) 
			for(j=0;j<n;j++)
				ynext[j] += xi[i][j];
		norm2 = 0;
		for(i=0;i<n;i++) 
		{
			ynext[i]/=n;
			norm2 += (ynext[i]-y[i])*(ynext[i]-y[i]);
		}
		norm2/=n;

		temp = y;
		y = ynext;
		ynext = temp;

		printf("objf: %f norm2: %f\n",mwisComputeObjFuncY( w, A, n, y),norm2);
	}

	for(i=0;i<n;i++) x[i] = (y[i]>=0.5);

	/* free memory */
	free(*xi);
	free(xi);
	free(*li);
	free(li);
	free(avgli);
	free(x0);
	free(x1);
	free(y);
	free(ynext);

	return x;
}
/* MWIS */
unsigned char* mwisGetX( MAT_TYPE* w, MAT_TYPE* A, int n, unsigned char* x)
{
	unsigned char option = 1;

	switch(option)
	{
		default:
			return mwisGetX2( w, A, n, x);
	}
}
/*
	
*/
#undef alpha